* A small capslock indicator for the Q40/Q60 v1.00 (c) Thierry Godefroy 2001.
* This software is open source.

          data      0
          section   CAPSMON

VERSION   equ       '1.10'

* CAPSMON thing definition block length and offset to the first variable in
* the thing linkage block:

THGDEF    equ       52

* Offset to the polled interrupt linkage in the thing linkage block. This is
* arranged so that the first variable is at offset 0 from the "base of device
* driver definition block" as passed in A3 when the polled interrupt routine
* is called.

POLLLINK  equ       THGDEF+8

* Offset (relative to A3) to the variables used by the polled interrupt:

MONITOR   equ       0                             Set this flag to monitor.
LED_STATE equ       1                             LED state indicator.
ONTICKS   equ       2                             Number of ticks on.
OFFTICKS  equ       4                             Number of ticks off.
COUNTER   equ       6                             Counter value for timing.

*******************************************************************************
* Initialization routine:

          MOVEQ     #1,D7                         Set the first init flag.
          BRA.S     INIT                          Initialize CAPSMON thing now.
CAPSLINK  MOVEQ     #0,D7                         Reset the "first init" flag.
INIT      MOVEQ     #0,D0                         | Get system info.
          TRAP      #1                            |
          MOVEQ     #-19,D0                       "Not implemented" error code.
          CMPI.L    #'SMSQ',(A0)                  Are we running under SMSQ(/E)?
          BNE.S     INIT_END                      If not then return with error.
          MOVE.B    $A7(A0),D2                    Get the machine type.
          ANDI.B    #$1F,D2                       Get rid of display type flags.
          CMPI.B    #$11,D2                       Is this a Q40/Q60 ?
          BNE.S     INIT_END                      If no then return with error.
          MOVEQ     #-2,D0                        "Invalid job Id" error code.
          TST.L     D1                            Is job 0 the calling job ?
          BNE.S     INIT_END                      If no, then return with error.
          BSR.S     INIT_THG                      Initialize the thing.
          BNE.S     INIT_END                      Return if error.
          TST.B     D7                            Is the "first init" flag set ?
          BEQ.S     INIT1                         If no then jump over...
          LEA       PROC_FN,A1                    PROC/FN table address.
          MOVEA.W   $110,A2                       | Link in the new PROCs/FNs.
          JSR       (A2)                          |
          BNE.S     INIT_END                      If error, return.
INIT1     LEA       COPYRIGHT,A1                  Get copyright message address.
          SUBA.L    A0,A0                         SBASIC #0 channel Id.
          MOVEA.W   $D0,A2                        | Print the copyright.
          JSR       (A2)                          |
INIT_END  RTS                                     Return to SBASIC.

*******************************************************************************
* CAPSMON thing initialization:

INIT_THG  MOVEQ     #0,D2                         | Reserves some space for the
          MOVEQ     #THGDEF+16,D1                 |  thing linkage block in the
          MOVEQ     #$18,D0                       |  common heap (owner is the
          TRAP      #1                            |  un-removable job 0).
          TST.L     D0                            Error ?
          BNE.S     INIT_END                      If yes, return immediately.
          MOVEA.L   A0,A3                         Save linkage block address.
          ADDQ.L    #8,A0                         Jump over system links.
          CLR.L     (A0)+                         Use default TH_FRFRE routine.
          CLR.L     (A0)+                         Use default TH_FRZAP routine.
          LEA       CAPS_THG,A1                   CAPSMON thing address.
          MOVE.L    A1,(A0)+                      Store thing address.
          CLR.L     (A0)+                         Use default TH_USE routine.
          CLR.L     (A0)+                         Use default TH_FREE routine.
          CLR.L     (A0)+                         Use default TH_FFREE routine.
          LEA       THGRMV,A1                     CAPSMON thing TH_REMOV addr.
          MOVE.L    A1,(A0)+                      Use my routine for TH_REMOV.
          CLR.W     (A0)+                         This thing is shareable.
          MOVE.L    #VERSION,(A0)+                Thing version ID.
          MOVE.W    #7,(A0)+                      Length of thing name.
          MOVE.L    #'CAPS',(A0)+                 | Name of thing.
          MOVE.L    #'MON ',(A0)+                 |
          BSR.S     SEARCH                        Search for THING vector.
          BNE.S     INITTHG1                      If error, deallocate linkage.
          MOVEA.L   A3,A1                         A1=thing linkage block addr.
          MOVEQ     #$26,D0                       | Link in my thing.
          JSR       (A4)                          |
          TST.L     D0                            Is there any error ?
          BEQ.S     INITTHG2                      If no, jump over...
INITTHG1  MOVE.L    D0,D4                         Save error code.
          MOVEA.L   A3,A0                         Thing linkage block address.
          MOVEQ     #$19,D0                       | Release common heap area
          TRAP      #1                            |  (thing linkage block).
          MOVE.L    D4,D0                         Restore error code.
          RTS                                     Return
INITTHG2  LEA       ENABLE,A2                     Get default from config block.
          MOVE.B    (A2),THGDEF+MONITOR(A3)       Set MONITOR flag.
          ST        THGDEF+LED_STATE(A3)          LED is on right now.
          LEA       TICKS_ON,A2                   Get default from config block.
          MOVE.W    (A2),THGDEF+ONTICKS(A3)       Set "on" ticks.
          LEA       TICKS_OFF,A2                  Get default from config block.
          MOVE.W    (A2),THGDEF+OFFTICKS(A3)      Set "off" ticks.
          LEA       POLLLINK(A3),A0               Address of polled int. link.
          LEA       POLLINT,A2                    Polled interrupt address.
          MOVE.L    A2,4(A0)                      Store it.
          MOVEQ     #$1C,D0                       | Link in the polled
          TRAP      #1                            |  interrupt routine.
          MOVEQ     #0,D0                         No error (above never fails).
          RTS                                     Return.

********************************************************************************
* CAPSMON_UNLINK (PROC), syntax:  CAPSMON_UNLINK
*
* This procedure unlinks the CAPSMON thing. Note that the SBASIC extensions are
* _not_ removed by this code.

CAPSUNLK  BSR.S     SEARCH                        Get the THING vector.
          BNE.S     EXIT                          If error then return.
          LEA       CAPSMON,A0                    Name of the thing to remove.
          MOVEQ     #$2A,D0                       | Remove the CAPSMON thing.
          JMP       (A4)                          |

********************************************************************************
* Routine searching for the "THING" vector:
*
* Call parameters   : none.
* ~~~~~~~~~~~~~~~
* Return parameters : A0.L = "THING" thing address (if D0=0)
* ~~~~~~~~~~~~~~~~~   A4.L = "TH_ENTRY" routine address
*                     D0.L = error code (0: no error, -19: not implemented),
*
* Modified registers: D0, A0, A4.
* ~~~~~~~~~~~~~~~~~~

SEARCH    MOVEM.L   D1/D2,-(SP)                   Save the context.
          MOVEQ     #0,D0                         | Get system info (we need
          TRAP      #1                            |  the system variable addr.).
          MOVEQ     #-19,D0                       "Not implemented" error code.
          LEA       $B8(A0),A0                    Things linked list pointer.
          MOVE.L    A0,D1                         Is it zero ?
          BEQ.S     SEARCH3                       If yes, return (no things).
SEARCH1   MOVE.L    (A0),D1                       D1=Thing linkage block addr.
          BEQ.S     SEARCH2                       If last thing then end.
          MOVEA.L   D1,A0                         | Else search for the next
          BRA.S     SEARCH1                       |  thing.
SEARCH2   MOVEA.L   16(A0),A0                     Pointer on the thing header.
          CMPI.L    #'THG%',(A0)                  Is it actually a thing ?
          BNE.S     SEARCH3                       If no then return with error.
          CMPI.L    #-1,4(A0)                     Is it the "THING" thing ?
          BNE.S     SEARCH3                       If no then return with error.
          MOVEA.L   8(A0),A4                      Get the TH_ENTRY vector.
          MOVEQ     #0,D0                         No error.
SEARCH3   MOVEM.L   (SP)+,D1/D2                   Restore the context.
          TST.L     D0                            Set the CCR.
EXIT      RTS                                     Return.

********************************************************************************
* This code is called when the CAPSMON thing is removed.

THGRMV    MOVEA.L   A1,A5                         Save thing linkage block addr.

* Unlink the polled interrupt task:

          LEA       POLLLINK(A5),A0               Link addr. of polled task.
          MOVEQ     #$1D,D0                       | Remove the CAPSMON thing
          TRAP      #1                            |  polled interrupt task.
          MOVE.B    #1,$FF000030                  Ensure the power LED is on.

* Finally we must return the thing linkage block to the common heap:

          MOVEA.L   A5,A0                         Thing linkage block address.
          MOVEQ     #$19,D0                       | De-allocate this block from
          TRAP      #1                            |  the common heap.
          MOVEQ     #0,D0                         No error.
          RTS                                     Return.

*******************************************************************************
* CAPSMON thing polled interrupt routine.

POLLINT   TST.B     (A3)                          Flag set ?
          BEQ.S     POLLINT1                      No, ensure LED is on...
          TST.L     ONTICKS(A3)                   Ticks on=ticks off=0 ?
          BNE.S     POLLINT0                      No, blinking LED monitor...
          TST.B     $88(A6)                       Test CAPSLOCK ?
          SNE       $FF000030                     LED state = CAPSLOCK state.
          SNE       LED_STATE(A3)                 And remember its state.
          RTS                                     Return.
POLLINT0  TST.B     $88(A6)                       CAPSLOCK on ?
          BNE.S     POLLINT2                      Yes, check counter...
POLLINT1  TST.B     LED_STATE(A3)                 Is the LED on ?
          BNE.S     POLLINT4                      If yes, return now.
          ST        $FF000030                     Turn the LED on...
          ST        LED_STATE(A3)                 and remember its state.
          MOVE.W    ONTICKS(A3),COUNTER(A3)       Preset the counter.
POLLINT4  RTS                                     Return.
POLLINT2  SUBQ.W    #1,COUNTER(A3)                Decrement counter.
          BPL.S     POLLINT4                      If still positive, return now.
          MOVE.W    ONTICKS(A3),D0                Ticks number for LED on.
          TST.B     LED_STATE(A3)                 Get the current LED state.
          SEQ       $FF000030                     Invert the LED state...
          SEQ       LED_STATE(A3)                 and remember it.
          BEQ.S     POLLINT3                      The LED is actually on now...
          MOVE.W    OFFTICKS(A3),D0               The LED is actually off now.
POLLINT3  MOVE.W    D0,COUNTER(A3)                Store number of ticks to go.
          RTS                                     Return

********************************************************************************
* CAPSMON (PROC), syntax: CAPSMON 1 or CAPSMON 0

CAPS_MON  MOVEA.W   $112,A2                       | Get an integer.
          JSR       (A2)                          |
          BNE.S     CAPSMON1                      If error, return now !
          MOVEQ     #-15,D0                       "Bad parameter" error code.
          SUBQ.W    #1,D3                         More than one integer ?
          BNE.S     CAPSMON1                      If yes, return with error.
          MOVE.W    0(A6,A1.L),D7                 Get the parameter value.
          BSR       SEARCH                        Get the THING vector.
          BNE.S     CAPSMON1                      If error then return.
          LEA       CAPSMON,A0                    Name of the thing to use.
          MOVE.L    #'OFF ',D2                    Name of the extension.
          TST.W     D7                            Was the parameter 0 ?
          BEQ.S     CAPSMON2                      Yes, so jump and do it...
          MOVE.L    #'ON  ',D2                    No, use 'ON' extension.
CAPSMON2  MOVEQ     #-1,D1                        User = calling job.
          MOVEQ     #$28,D0                       | Use the CAPSMON thing.
          JSR       (A4)                          |
          TST.L     D0                            Error ?
          BNE.S     CAPSMON1                      If yes then return now !
          JSR       $18(A1)                       Call the extension thing.
          MOVEQ     #$29,D0                       | Free the CAPSMON thing.
          JSR       (A4)                          |
CAPSMON1  RTS                                     Return to SBASIC.

* CAPSMON_TICKS (PROC), syntax: CAPSMON_TICKS on_ticks,off_ticks

CAPSTICKS MOVEA.W   $112,A2                       | Get integers.
          JSR       (A2)                          |
          BNE.S     CAPSMON1                      If error, return now !
          MOVEQ     #-15,D0                       "Bad parameter" error code.
          SUBQ.W    #2,D3                         Exactly two integers ?
          BNE.S     CAPSMON1                      If no, return with error.
          MOVEA.L   A1,A5                         Save the arithmetic stack ptr.
          BSR       SEARCH                        Get the THING vector.
          BNE.S     CAPSMON1                      If error then return.
          LEA       CAPSMON,A0                    Name of the thing to use.
          MOVE.L    #'TICK',D2                    No, use 'TICK' extension.
          MOVEQ     #-1,D1                        User = calling job.
          MOVEQ     #$28,D0                       | Use the CAPSMON thing.
          JSR       (A4)                          |
          TST.L     D0                            Error ?
          BNE.S     CAPSMON1                      If yes then return now !
          MOVE.W    0(A6,A5.L),D1                 Get the "on_ticks" value.
          MOVE.W    2(A6,A5.L),D2                 Get the "off_ticks" value.
          JSR       $18(A1)                       Call the extension thing.
          MOVEQ     #-1,D1                        User = calling job.
          MOVEQ     #$29,D0                       | Free the CAPSMON thing.
          JMP       (A4)                          |

*******************************************************************************
* Extension thing:

CAPS_THG

* 'ON  ' extension: sets the monitor flag to "on" (no parameter, always
*                   succeed.

CAPS_THG1 dc.b      'THG%'                        Thing header marker.
          dc.l      $01000003                     Extension thing with list.
          dc.l      CAPS_THG2-CAPS_THG1           Pointer to next extension.
          dc.b      'ON  '                        Id for the first extension.
          dc.l      0                             No parameter definition list.
          dc.l      0                             No parameter description list.
          ST        THGDEF+MONITOR(A2)            Enable the monitoring.
          MOVEQ     #0,D0                         No error.
          RTS                                     Return.

* 'OFF ' extension: sets the monitor flag to "off" (no parameter, always
*                   succeed.

CAPS_THG2 dc.b      'THG%'                        Thing header marker.
          dc.l      $01000003                     Extension thing with list.
          dc.l      CAPS_THG3-CAPS_THG2           Pointer to next extension.
          dc.b      'OFF '                        Id for the first extension.
          dc.l      0                             No parameter definition list.
          dc.l      0                             No parameter description list.
          SF        THGDEF+MONITOR(A2)            Disable the monitoring.
          MOVEQ     #0,D0                         No error.
          RTS                                     Return.

* 'TICK' extension: sets the number of LED "on" ticks (0 to 32767 in D1), and
*                   the number of "off" ticks (0 to 32767 in D2). Always succeed
*                   (beware: the parameter values are not checked).

CAPS_THG3 dc.b      'THG%'                        Thing header marker.
          dc.l      $01000003                     Extension thing with list.
          dc.l      0                             No more extension.
          dc.b      'TICK'                        Id for the first extension.
          dc.l      0                             No parameter definition list.
          dc.l      0                             No parameter description list.
          MOVE.W    D1,THGDEF+ONTICKS(A2)         Store the "on" ticks.
          MOVE.W    D2,THGDEF+OFFTICKS(A2)        Store the "off" ticks.
          CLR.W     THGDEF+COUNTER(A2)            Reset the counter.
          MOVEQ     #0,D0                         No error.
          RTS                                     Return.

********************************************************************************
* Data:

* PROCedures and FuNctions definition table:

PROC_FN   dc.w      7
          dc.w      CAPSLINK-*
          dc.b      12,'CAPSMON_LINK',0
          dc.w      CAPSUNLK-*
          dc.b      14,'CAPSMON_UNLINK',0
          dc.w      CAPS_MON-*
          dc.b      7,'CAPSMON'
          dc.w      CAPSTICKS-*
          dc.b      13,'CAPSMON_TICKS'
          dc.w      0
          dc.w      0
          dc.w      0

* Copyright message:

COPYRIGHT dc.w      COPYREND-*-1
          dc.b      'CAPSMON (CAPSLOCK indicator) for Q40/Q60 v'
          dc.l      VERSION
          dc.b      ' (c) 2001 Thierry GODEFROY.'
COPYREND  dc.b      10
          dc.w      0

* Thing name:

CAPSMON   dc.w      7
          dc.b      'CAPSMON',0

* Configuration block:

          dc.b      '<<QCFX>>'                    CONFIG block Id.
          dc.b      '01'                          Required config level.
          dc.w      7                             Software name length.
          dc.b      'CAPSMON',0                   Software name.
          dc.w      4                             Version string length.
          dc.l      VERSION                       Version string.

          dc.b      4                             Item type  = code.
          dc.b      0                             Selection key (none).
          dc.w      ENABLE-*                      Ptr to code value.
          dc.w      0                             Ptr to pre-processing code.
          dc.w      0                             Ptr to post-processing code.
          dc.w      ENAB_TEXT-*                   Ptr to item description.
          dc.w      ENAB_ATTR-*                   Ptr to item attributes.

          dc.b      10                            Item type=numeric (word).
          dc.b      0                             Selection key (none).
          dc.w      TICKS_ON-*                    Ptr to numeric value.
          dc.w      0                             Ptr to pre-processing code.
          dc.w      0                             Ptr to post-processing code.
          dc.w      ON_TEXT-*                     Ptr to item description.
          dc.w      VAL_ATTR-*                    Ptr to item attributes.

          dc.b      10                            Item type=numeric (word).
          dc.b      0                             Selection key (none).
          dc.w      TICKS_OFF-*                   Ptr to numeric value.
          dc.w      0                             Ptr to pre-processing code.
          dc.w      0                             Ptr to post-processing code.
          dc.w      OFF_TEXT-*                    Ptr to item description.
          dc.w      VAL_ATTR-*                    Ptr to item attributes.

          dc.w      -1                            End of configurable items.

ENABLE    dc.b      1                             Default value (enabled).
          dc.b      0                             Aligment byte.
ENAB_TEXT dc.w      26                            | Invite string.
          dc.b      'CAPSMON enabled by default'  |
ENAB_ATTR dc.b      0                             1st possible value.
          dc.b      0                             No selection key.
          dc.w      2                             | associated answer.
          dc.b      'no'                          |
          dc.b      1                             2nd possible value.
          dc.b      0                             No selection key.
          dc.w      3                             | associated answer.
          dc.b      'yes',0                       |
          dc.w      -1                            End of possible code values.

TICKS_ON  dc.w      30                            Default numeric value.
ON_TEXT   dc.w      26                            | Invite string.
          dc.b      'Number of ticks for LED on'  |

TICKS_OFF dc.w      30                            Default numeric value.
OFF_TEXT  dc.w      27                            |
          dc.b      'Number of ticks for LED'     | Invite string.
          dc.b      ' off',0                      |

VAL_ATTR  dc.w      0                             Minimum value.
          dc.w      32767                         Maximum value.

          end
